
/* ===================  AnyDBM  ===============================
Copyright (C) 1998, WebThing Ltd
Author: Nick Kew

This library is free software; you can redistribute it and/or
modify it under the terms of the GNU Library General Public
License as published by the Free Software Foundation; either
version 2 of the License, or (at your option) any later version.

This library is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
Library General Public License for more details.

You should have received a copy of the GNU Library General Public
License along with this library; if not, write to the
Free Software Foundation, Inc., 59 Temple Place - Suite 330,
Boston, MA  02111-1307, USA.

-------------
Author/Copyright Holder Information: please see the README file and
the AnyDBM webpages at <URL:http://www.webthing.com/software/AnyDBM/>

============================================================ */
#ifndef RDBM_CC_H
#define RDBM_CC_H

#include <rpc.h>
#include "rdbm.h"
#include "rdbm.const"
#include "BASE_DBM.h"

template <class T>
class rpc_buf {
  friend class RDBM<T> ;
private:
  struct rdbm_key_val_t _data ;
  rpc_buf() ;					// set ident
  rdbm_ident_t* id() { return &_data.id ; }
  int set(const char*, const T&) ;
  rdbm_key_t* key() { return (rdbm_key_t*) &_data ; }
  rdbm_key_val_t* data() { return &_data ; }
} ;


template <class T>
class RDBM : public BASE_DBM<T> {
protected:
  void* dbm_d() const { return SERVER ; }
  int file_d() const { return 0 ; }
  bool has_fd() const { return false ; }
/* RDBM delegates locking to the server */
  bool always_locked() const { return true ; }
  bool lock(short) { return true ; }
  bool unlock() { return true ; }
private:
  CLIENT* SERVER ;
  int _errnum ;					// internal errnum
  const char* _val ;				// that-which-was retrieved
  rpc_buf<T> _buf ;
  bool need_sync ;

  void handle(int* status) ;			// set _val and _errnum
  void handle(char** status) ;
public:
  RDBM( const char* host ="localhost",
	const int prog =RDBM_P,
	const int vers =RDBM_V) ;		// init and connect
  int reconnect(const char* host ="localhost",
	const int prog =RDBM_P,
	const int vers =RDBM_V) ;		// make connection.

  virtual ~RDBM() ;				// break connection
  void clear_err(void) { _errnum = 0 ; }

/* these return errnum. The last two set _val to the char* string fetched */
  int sync() ;
  int put_new(const char* key, const T& val) ;	// enter (no overwrite)
  int put(const char* key, const T& val) ;	// enter
  int del(const char* key) ;			// delete
  int fetch(const char* key) ;			// lookup
  int nextkey(const char* key) ;		// nextkey
  int firstkey() { return nextkey("") ;	}	// yeah OK

/* When either of these returns zero you should check errnum */
  T get(const char* key)			// lookup
	{ if (fetch(key)) return (const char*)0 ; else return _val ; }
  const char* seq(const char* key)		// nextkey
	{ if (nextkey(key)) return 0 ; else return _val ; }

  int errnum() const { return _errnum ; } 
  const char* errmsg() const ;
  const T lastval() const { return _val ; }
} ;

#include <iostream.h>
#include "rdbm.const"
#include "rdbm.h"

template <class T>
inline const char* RDBM<T>::errmsg() const {
  switch (_errnum) {
  case 0:
    return "" ;
  case RDBM_RPC_FAILED:
    return "RPC Error talking to rdbmd" ;
  case RDBM_NO_SERVER:
    return "Couldn't connect to rdbmd" ;
  case RDBM_NO_KEY:
    return "Invalid RDBM call - no key" ;
  case RDBM_KEY_TOO_LONG:
    return "Invalid RDBM call - key too long (change this in rdbm.const)" ;
  case RDBM_VAL_TOO_LONG:
    return "Invalid RDBM call - data too long (change this in rdbm.const)" ;
  case RDBM_DBM_FAILED:
    return "DBM operation failed" ;
  case RDBM_DBM_LOCKED:
    return "DBM locked by another process (please try again)" ;
  case RDBM_NO_DATA:
    return "No data found at this point in DBM" ;
  }
  return "Aaargh... a bug!   Invalid errnum." ;	// shouldn't reach here
}
template <class T>
inline void RDBM<T>::handle(int* status) {
  if (status)
    switch (*status) {
      case -1:
	_errnum = RDBM_DBM_LOCKED ;
	break ;
      case 0:
	_errnum = RDBM_DBM_FAILED ;
	break ;
      default:
	_errnum = 0 ;
	break ;
    }
  else
    _errnum = RDBM_RPC_FAILED ;
}
template <class T>
inline void RDBM<T>::handle(char** status) {
  _val = 0 ;
  if (status)
    switch ((int)status[0]) {
      case -1:
	_errnum = RDBM_DBM_LOCKED ;
	break ;
      case 0:
	_errnum = RDBM_DBM_FAILED ;
	break ;
      default:
	_val = *status ;
	if ( *_val )
	  _errnum = 0 ;
	else
	  _errnum = RDBM_NO_DATA ;
	break ;
    }
  else
    _errnum = RDBM_RPC_FAILED ;
}
template <class T>
inline int RDBM<T>::put(const char* key, const T& val) {
  if ( _errnum < 0 ) return _errnum ;
  need_sync = true ;
  if (_errnum = _buf.set(key, val), !_errnum)
    handle(rdbm_put_1(_buf.data(), SERVER)) ;
  return _errnum ;
}
template <class T>
inline int RDBM<T>::put_new(const char* key, const T& val) {
  if ( _errnum < 0 ) return _errnum ;
  need_sync = true ;
  if (_errnum = _buf.set(key, val), !_errnum)
    handle(rdbm_put_new_1(_buf.data(), SERVER)) ;
  return _errnum ;
}
template <class T>
inline int RDBM<T>::del(const char* key) {
  if ( _errnum < 0 ) return _errnum ;
  need_sync = true ;
  if (_errnum = _buf.set(key, (_val = "", _val)), !_errnum)
    handle(rdbm_del_1(_buf.key(), SERVER)) ;
  return _errnum ;
}
template <class T>
inline int RDBM<T>::fetch(const char* key) {
  if ( _errnum < 0 ) return _errnum ;
  if (_errnum = _buf.set(key, _val), !_errnum)
    handle(rdbm_get_1(_buf.key(), SERVER)) ;
  return _errnum ;
}
template <class T>
inline int RDBM<T>::nextkey(const char* key) {
  if ( _errnum < 0 ) return _errnum ;
  if (_errnum = _buf.set(key, _val), !_errnum)
    handle(rdbm_seq_1(_buf.key(), SERVER)) ;
  return _errnum ;
}
template <class T>
inline int RDBM<T>::sync() {
  if ( _errnum < 0 ) return _errnum ;
  handle(rdbm_sync_1(_buf.id(), SERVER)) ;
  need_sync = false ;
  return _errnum ;
}

/* to assign a session key .. not for authentication */
template <class T>
inline rpc_buf<T>::rpc_buf() {
  _data.id.pid = getpid() ;
  _data.id.host = gethostid() ;
}
template <class T>
inline int rpc_buf<T>::set(const char* key, const T& _val) {
  if ( ! key )
    return RDBM_NO_KEY ;
  if ( strlen(key) > rdbm_max_key )
    return RDBM_KEY_TOO_LONG ;
  strcpy(_data.key, key) ;

  const char* val = (const char*) _val ;
  if (val) {
    if ( strlen(val) > rdbm_max_data )
      return RDBM_VAL_TOO_LONG ;
    strcpy(_data.val, val) ;
  }
  return 0 ;
}
template <class T>
inline int RDBM<T>::reconnect(const char* host, const int p, const int v) {
  if (SERVER) {
    if (need_sync)
	sync() ;
    clnt_destroy(SERVER) ;
    SERVER = 0 ;
  }
  SERVER = clnt_create(host, p, v, "tcp") ;
  if ( !SERVER )
    _errnum = RDBM_NO_SERVER ;
  else
    _errnum = 0 ;
  return _errnum ;
}
template <class T>
inline RDBM<T>::RDBM(const char* host, const int p, const int v) : SERVER(0) {
  need_sync = false ;
  _val = (const char*)0 ;
  reconnect(host, p, v) ;
}
template <class T>
inline RDBM<T>::~RDBM() {
  if (need_sync)
    sync() ;
  if (SERVER) {
    clnt_destroy(SERVER) ;
    SERVER = 0 ;
  }
}

#endif
