/* ====================================================================
 * Copyright (c) 1995-1998 The Apache Group.  All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 *
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer. 
 *
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in
 *    the documentation and/or other materials provided with the
 *    distribution.
 *
 * 3. All advertising materials mentioning features or use of this
 *    software must display the following acknowledgment:
 *    "This product includes software developed by the Apache Group
 *    for use in the Apache HTTP server project (http://www.apache.org/)."
 *
 * 4. The names "Apache Server" and "Apache Group" must not be used to
 *    endorse or promote products derived from this software without
 *    prior written permission. For written permission, please contact
 *    apache@apache.org.
 *
 * 5. Redistributions of any form whatsoever must retain the following
 *    acknowledgment:
 *    "This product includes software developed by the Apache Group
 *    for use in the Apache HTTP server project (http://www.apache.org/)."
 *
 * THIS SOFTWARE IS PROVIDED BY THE APACHE GROUP ``AS IS'' AND ANY
 * EXPRESSED OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR
 * PURPOSE ARE DISCLAIMED.  IN NO EVENT SHALL THE APACHE GROUP OR
 * ITS CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
 * SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT
 * NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
 * LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT,
 * STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
 * ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED
 * OF THE POSSIBILITY OF SUCH DAMAGE.
 * ====================================================================
 *
 * This software consists of voluntary contributions made by many
 * individuals on behalf of the Apache Group and was originally based
 * on public domain software written at the National Center for
 * Supercomputing Applications, University of Illinois, Urbana-Champaign.
 * For more information on the Apache Group and the Apache HTTP server
 * project, please see <http://www.apache.org/>.
 *
 */

/* This software is contributed to the Apache Group by WebThing Ltd,
   <URL:http://www.webthing.com/>
*/

/*
	mod_auth_rdbm:	networked authentication using an AnyDBM server

	Copyright (c) WebThing Ltd, 1998.
	Author Nick Kew	<URL:mailto:nick@webthing.com>

	This is free software, under the same License and No Warranty
	conditions as the Apache Webserver itself.

	See the README file for a description, and the AnyDBM web pages
	<URL:http://www.webthing.com/software/AnyDBM/> for updates.
*/

#include "httpd.h"
#include "http_config.h"
#include "http_core.h"
#include "http_log.h"
#include "http_protocol.h"
#include <AnyDBM/RDBM_C.h>
#include <AnyDBM/rdbm.h>

/*
 * Module definition information - the part between the -START and -END
 * lines below is used by Configure. This could be stored in a separate
 * instead.
 *
 * This is for sensible systems like Linux:
 *  LIBS="$LIBS -lrdbm"
 * You might need something more - e.g. on Solaris you would need:
 *  LIBS="$LIBS -lrdbm -lnsl"
 *
 * MODULE-DEFINITION-START
 * Name: rdbm_auth_module
 * ConfigStart
    LIBS="$LIBS -lrdbm"
 * ConfigEnd
 * MODULE-DEFINITION-END
 */

typedef struct {

    char *auth_rdbm_host;
    int auth_rdbm_prog;
    int auth_rdbm_vers;
    int auth_rdbmauthoritative;

} rdbm_auth_config_rec;

static void *create_rdbm_auth_dir_config(pool *p, char *d)
{
    rdbm_auth_config_rec *sec
    = (rdbm_auth_config_rec *) ap_pcalloc(p, sizeof(rdbm_auth_config_rec));

    sec->auth_rdbm_host = "localhost" ;
    sec->auth_rdbm_prog = RDBM_P;
    sec->auth_rdbm_vers = RDBM_V;
    sec->auth_rdbmauthoritative = 1;	/* fortress is secure by default */

    return sec;
}


static command_rec rdbm_auth_cmds[] =
{
    {"AuthRDBMHost", ap_set_string_slot,
     (void *) XtOffsetOf(rdbm_auth_config_rec, auth_rdbm_host),
     OR_AUTHCFG, TAKE1, NULL},
    {"AuthRDBMAuthoritative", ap_set_flag_slot,
     (void *) XtOffsetOf(rdbm_auth_config_rec, auth_rdbmauthoritative),
     OR_AUTHCFG, FLAG, "Set to 'no' to allow access control to be passed along to lower modules, if the UserID is not known in this module"},
    {NULL}
};

module rdbm_auth_module;

static const char *get_rdbm_pw(request_rec *r, char *user,
	char *host, int prog, int vers)
{
    rdbm_fd handle ;
    const char *pw = NULL;

    handle = rdbm_open(host,prog,vers) ;
    if (!handle) {
	ap_log_error(APLOG_MARK, APLOG_ERR, r->server,
		    "Bad RDBM", host);
	return NULL;
    }
    if (rdbm_errnum(handle)) {
	ap_log_error(APLOG_MARK, APLOG_ERR, r->server,
		    rdbm_errmsg(handle), host);
	return NULL;
    }

    switch (rdbm_fetch(handle, user) ) {
	case 0: pw = rdbm_lastval(handle) ; /* fallsthrough */
	default: break ;		/* hmmm.... */
    }

    rdbm_close(handle);
    return pw;
}

/* We do something strange with the group file.  If the group file
 * contains any : we assume the format is
 *      key=username value=":"groupname [":"anything here is ignored]
 * otherwise we now (0.8.14+) assume that the format is
 *      key=username value=groupname
 * The first allows the password and group files to be the same 
 * physical DBM file;   key=username value=password":"groupname[":"anything]
 *
 * mark@telescope.org, 22Sep95
 */

static char *get_rdbm_grp(request_rec *r, char *user,
	char* host, int prog, int vers)
{
    const char *grp_data = get_rdbm_pw(r, user, host, prog, vers);
    char *grp_colon;
    char *grp_colon2;

    if (grp_data == NULL)
	return NULL;

    if ((grp_colon = strchr(grp_data, ':')) != NULL) {
	grp_colon2 = strchr(++grp_colon, ':');
	if (grp_colon2)
	    *grp_colon2 = '\0';
	return grp_colon;
    }
    return grp_data;
}

static int rdbm_authenticate_basic_user(request_rec *r)
{
    rdbm_auth_config_rec *sec =
    (rdbm_auth_config_rec *) ap_get_module_config(r->per_dir_config,
					      &rdbm_auth_module);
    conn_rec *c = r->connection;
    char *sent_pw, *colon_pw;
    const char* real_pw ;
    int res;

    if ((res = ap_get_basic_auth_pw(r, &sent_pw)))
	return res;

    if (!(real_pw = get_rdbm_pw(r, c->user, sec->auth_rdbm_host,
	sec->auth_rdbm_prog, sec->auth_rdbm_vers))) {
	if (!(sec->auth_rdbmauthoritative))
	    return DECLINED;
	ap_log_error(APLOG_MARK, APLOG_NOERRNO|APLOG_ERR, r->server,
		    "RDBM user %s not found: %s", c->user, r->filename);
	ap_note_basic_auth_failure(r);
	return AUTH_REQUIRED;
    }
    /* Password is up to first : if exists */
    colon_pw = strchr(real_pw, ':');
    if (colon_pw)
	*colon_pw = '\0';
    /* anyone know where the prototype for crypt is? */
    if (strcmp(real_pw, (char *) crypt(sent_pw, real_pw))) {
	ap_log_error(APLOG_MARK, APLOG_NOERRNO|APLOG_ERR, r->server,
		    "user %s: password mismatch: %s", c->user, r->uri);
	ap_note_basic_auth_failure(r);
	return AUTH_REQUIRED;
    }
    return OK;
}

/* Checking ID */

static int rdbm_check_auth(request_rec *r)
{
    rdbm_auth_config_rec *sec =
    (rdbm_auth_config_rec *) ap_get_module_config(r->per_dir_config,
					      &rdbm_auth_module);
    char *user = r->connection->user;
    int m = r->method_number;

    array_header *reqs_arr = ap_requires(r);
    require_line *reqs = reqs_arr ? (require_line *) reqs_arr->elts : NULL;

    register int x;
    const char *t;
    char *w;

    if (!reqs_arr)
	return DECLINED;

    for (x = 0; x < reqs_arr->nelts; x++) {

	if (!(reqs[x].method_mask & (1 << m)))
	    continue;

	t = reqs[x].requirement;
	w = ap_getword(r->pool, &t, ' ');

	if (!strcmp(w, "group")) {
	    const char *orig_groups, *groups;
	    char *v;

	    if (!(groups = get_rdbm_grp(r, user, sec->auth_rdbm_host,
	sec->auth_rdbm_prog, sec->auth_rdbm_vers))) {
		if (!(sec->auth_rdbmauthoritative))
		    return DECLINED;
		ap_log_error(APLOG_MARK, APLOG_NOERRNO|APLOG_ERR, r->server,
			    "user %s not in RDBM group file on %s",
			    user, sec->auth_rdbm_host);
		ap_note_basic_auth_failure(r);
		return AUTH_REQUIRED;
	    }
	    orig_groups = groups;
	    while (t[0]) {
		w = ap_getword(r->pool, &t, ' ');
		groups = orig_groups;
		while (groups[0]) {
		    v = ap_getword(r->pool, &groups, ',');
		    if (!strcmp(v, w))
			return OK;
		}
	    }
	    ap_log_error(APLOG_MARK, APLOG_NOERRNO|APLOG_ERR, r->server,
			"user %s not in right group: %s",
			user, r->filename);
	    ap_note_basic_auth_failure(r);
	    return AUTH_REQUIRED;
	}
    }

    return DECLINED;
}


module rdbm_auth_module =
{
    STANDARD_MODULE_STUFF,
    NULL,			/* initializer */
    create_rdbm_auth_dir_config,	/* dir config creater */
    NULL,			/* dir merger --- default is to override */
    NULL,			/* server config */
    NULL,			/* merge server config */
    rdbm_auth_cmds,		/* command table */
    NULL,			/* handlers */
    NULL,			/* filename translation */
    rdbm_authenticate_basic_user,	/* check_user_id */
    rdbm_check_auth,		/* check auth */
    NULL,			/* check access */
    NULL,			/* type_checker */
    NULL,			/* fixups */
    NULL,			/* logger */
    NULL,			/* header parser */
    NULL,			/* child_init */
    NULL,			/* child_exit */
    NULL			/* post read-request */
};
